local LootBuddy = LibStub("AceAddon-3.0"):NewAddon("LootBuddy", "AceConsole-3.0", "AceEvent-3.0", "AceConsole-3.0");
local L = LibStub('AceLocale-3.0'):GetLocale('LootBuddy', true)

local options = {
	name = "LootBuddy",
	handler = LootBuddy,
	type = "group",
	args = {
		c = {
			name = "Settings",
			type = "group",
			inline = true,
			args = {
				fenabled = {
					name = 'Trashing Enabled',
					desc = "Enable LootBuddy filtering on loot items.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 1,
				},
				debug_mode = {
					name = 'Debug Mode',
					desc = "Loot will not be trashed, you'll receive notices.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 2,
				},
				silent_mode = {
					name = 'Silent Mode',
					desc = "Trashing chat notices.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 2,
				},
				cfilters = {
					name = 'Current Filters - Malformed filters are automatically removed',
					type = 'input',
					multiline = 14,
					width = 'full',
					desc = 'add, remove and edit filters, one filter per line.',
					set = 'SetFilters',
					get = 'GetFilters',
					order = 3,
				},
				safelist = {
					name = 'Safe List - Items in this list will never be removed',
					type = 'input',
					multiline = 10,
					width = 'full',
					desc = 'comma delimit item names.',
					set = 'SetSafeList',
					get = 'GetSafeList',
					order = 4,
				},
				help_label = {
					name = L["TagHelp"],
					type = 'description',
					order = 5,
					width = 'full',
				},
			}
		}
	}
}

local defaults = { 
	char = {
		fenabled=true,
		debug_mode=true,
		silent_mode=false,
		safelist="",
		cfilters="",
		dev_debug=false,
	},
}
local loot, bags, filters, trash, safe = {}, {}, {}, {}, {}
local bag_updates = 0

function LootBuddy:OnInitialize()
	LootBuddy.db = LibStub("AceDB-3.0"):New("LootBuddyFiltersDB", defaults)
	LootBuddy.version = GetAddOnMetadata("LootBuddy", "Version");
	LibStub("AceConfig-3.0"):RegisterOptionsTable("LootBuddy", options)
	LootBuddy.optionsDialog = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("LootBuddy")
	LootBuddy:SetFilters('', LootBuddy.db.char.cfilters)
	LootBuddy:SetSafeList('', LootBuddy.db.char.safelist);
end

function LootBuddy:OnEnable()
	print_str(": Version " .. LootBuddy.version)
	LootBuddy:RegisterEvent("BAG_UPDATE")
	LootBuddy:RegisterEvent("LOOT_OPENED")
	LootBuddy:RegisterChatCommand("lootbuddy", "OnSlashCommand")
	LootBuddy:RegisterChatCommand("lb", "OnSlashCommand")
	foodList = foodList or {}
	waterList = waterList or {}
end

function LootBuddy:OnDisable()
	LootBuddy:UnregisterEvent("BAG_UPDATE")
	LootBuddy:UnregisterEvent("LOOT_OPENED")
	LootBuddy:UnregisterChatCommand("lootbuddy")
	LootBuddy:UnregisterChatCommand("lb")
end

function LootBuddy:OnSlashCommand(input)
	local command, rest = input:match("^(%S*)%s*(.-)$");
	-- Any leading non-whitespace is captured into command;
	-- the rest (minus leading whitespace) is captured into rest.
	if command == "enable" then
		if LootBuddy.db.char.fenabled == false then
			LootBuddy.db.char.fenabled = true
			LootBuddy:RegisterEvent("BAG_UPDATE")
			LootBuddy:RegisterEvent("LOOT_OPENED")
			print_str(': enabled.')
		else
			print_str(': is already enabled.')
		end
	elseif command == "disable" then
		if LootBuddy.db.char.fenabled == true then
			LootBuddy.db.char.fenabled = false
			LootBuddy:UnregisterEvent("BAG_UPDATE")
			LootBuddy:UnregisterEvent("LOOT_OPENED")
			print_str(': disabled.')
		else
			print_str(': is already disabled.')
		end
	elseif command == "config" then
	-- Handle removing of the contents of rest... to something.   
		InterfaceOptionsFrame_OpenToCategory(LootBuddy.optionsDialog)
	elseif commend == "silent" then
		LootBuddy.db.char.silent_mode = true
	else
	-- If not handled above, display some sort of help message
		print_str(" Syntax: /lootbuddy config, /lootbuddy enable, /lootbuddy disable");
	end
end

function LootBuddy:LOOT_OPENED(event, auto)
	loot, bags = {}, {}
	if LootBuddy.db.char.fenabled and table.getn(filters) > 0 then
		for i=1,GetNumLootItems() do
			local lootIcon, lootName, lootQuantity, rarity, locked = GetLootSlotInfo(i)
			if lootQuantity ~= nil then
				if lootQuantity > 0 then
					table.insert(loot, {lootName, false})
				end
			end
		end
	end
end

function LootBuddy:BAG_UPDATE(event, bag)
	if LootBuddy.db.char.fenabled and table.getn(filters) > 0 then
		if table.getn(loot) > 0 then
			if in_table(bags, bag) ~= true then
				table.insert(bags, bag)
			end
			if (bag_updates+1) / 2 == table.getn(loot) then
				filter_loot()
				bag_updates = 0
			else
				bag_updates = bag_updates + 1
			end
		end
	end
end

function filter_loot()
	if LootBuddy.db.char.fenabled and table.getn(filters) > 0 then
		local name, link, total_slots
		local llen = table.getn(loot)
		for i=1,table.getn(loot) do
			for bag = 0, 4 do
				if in_table(bags, bag) == true then
					if loot[i][2] ~= true then
						total_slots = GetContainerNumSlots(bag);
						for slot = 1, total_slots do
							link = GetContainerItemLink(bag, slot);
							if link ~= nil then
								name = GetItemInfo(link)
								if name == loot[i][1] then
									onItem(bag, slot, link)
									loot[i][2] = true
									break
								end
							end
						end
					end
				end
			end
		end
		loot, bags = {}, {}
	end
end

function onItem(bag, slot, link)
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link)
	local stackValue = itemStackCount * itemSellPrice
	local itemColor = string.match(itemLink, "|?c?f?f?(%x*)|")
	if table.getn(filters) > 0 then
		if itemType ~= "Quest" and itemRarity < 4 and in_table(safe, itemName) == false then
			for mi,mf in pairs(filters) do
				local flen = table.getn(mf)
				local m = 0
				for index,filter in pairs(mf) do
					if filter[1] == "name" then
						if filter[3] == itemName then
							m = m + 1
						end
					elseif filter[1] == "type" then
						if LootBuddy.db.char.dev_debug then
							print_str(' type: ' .. itemType .. ' looking for: ' .. filter[3])
						end
						if filter[2] == "=" then
							if filter[3] == "Miscellaneous" then
								if LootBuddy.db.char.dev_debug then
									print_str('itemColor: ' .. itemColor)
								end
								if itemColor == "9d9d9d" then
									m = m + 1
									if LootBuddy.db.char.dev_debug then
										print_str('itemColor match ' .. m)
									end
								end
							elseif filter[3] == "food" and itemType == "Consumable" then
								local itemId = GetContainerItemID(bag, slot);
								if in_table(foodList, itemId) then
									m = m + 1
								end
							elseif filter[3] == "water" and itemType == "Consumable" then
								local itemId = GetContainerItemID(bag, slot);
								if in_table(waterList, itemId) then
									m = m + 1
								end
							elseif filter[3] == itemType then
								m = m + 1
							end
						elseif filter[2] == "!=" and filter[3] ~= itemType then
							m = m + 1
						end
					elseif filter[1] == "subtype" then
						if filter[2] == "=" and filter[3] == itemSubType then
							if LootBuddy.db.char.dev_debug then
								print_str('subtype found: ' .. itemSubType .. ' looking for subtype: ' .. filter[3])
							end
							m = m + 1
							if LootBuddy.db.char.dev_debug then
								print_str('subtype match ' .. m)
							end
						elseif filter[2] == "!=" and filter[3] ~= itemSubType then
							m = m + 1
						end
					elseif filter[1] == "quality" then
						if LootBuddy.db.char.dev_debug then
							print_str('quality found: ' .. itemRarity .. ' looking for quality: ' .. filter[3])
						end
						if filter[2] == "=" and tonumber(filter[3]) == itemRarity then
							m = m + 1
							if LootBuddy.db.char.dev_debug then
								print_str('quality match ' .. m)
							end
						elseif filter[2] == "!=" and tonumber(filter[3]) ~= itemRarity then
							m = m + 1
						end
					elseif filter[1] == "value" and itemSellPrice < tonumber(filter[3]) then
						m = m + 1
					elseif filter[1] == "level" and itemLevel < tonumber(filter[3]) then
						m = m + 1
					end
				end
				if LootBuddy.db.char.dev_debug then
					print_str('needed: ' .. flen .. ' found ' .. m)
				end
				if m == flen then
					if LootBuddy.db.char.debug_mode then
						print_str(': ' .. link .. ' marked to trash, no action taken - debug active.')
					else
						DeleteContainerItem(bag, slot, link)
					end
				end
			end
		end
	end
end

function DeleteContainerItem(bag, slot, link, filter)
	PickupContainerItem(bag, slot);
	if CursorHasItem() then
		DeleteCursorItem();
	end
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link);
	if LootBuddy.db.char.silent_mode ~= true then
		print_str(': ' .. link .. ' trashed.')
	end
end

function LootBuddy:GetOption(info)
	return LootBuddy.db.char[info[#info]]
end

function LootBuddy:SetOption(info, input)
	LootBuddy.db.char[info[#info]] = input
end

function LootBuddy:GetFilters(info)
	return LootBuddy.db.char.cfilters
end

function LootBuddy:SetFilters(info, input)
	filters = {}
	LootBuddy.db.char.cfilters = "" -- input
	if input ~= '' then
		local new_filter = {}
		local fstring = split_str(input, "[^\r\n]+", false);
		--if table.getn(fstring) > 0 then
			for i,f in ipairs(fstring) do
				new_filter = {}
				for w in f:gmatch("[%a]+%s[(=)(!=)(<)]+%s\"[%a%d%s]+[\"$]") do
					if LootBuddy.db.char.dev_debug then
						print_str(': matched filter found: ' .. w)
					end
					table.insert(new_filter, w)
				end
				if table.getn(new_filter) > 0 then
					table.insert(filters, new_filter)
					LootBuddy.db.char.cfilters = LootBuddy.db.char.cfilters .. table.concat(new_filter, " and ") .. "\r"
				end 
			end
			if LootBuddy.db.char.dev_debug then
				print_str('filters found: ' .. table.getn(filters))
			end
			if table.getn( filters ) > 0 then
				for mi,mf in pairs(filters) do
					if LootBuddy.db.char.dev_debug then
					end
					for index,fval in pairs(mf) do
						if LootBuddy.db.char.dev_debug then
							print_str('breaking down filter')
						end
						local pattern = string.explode(fval:gsub("([%a]+)%s([(=)(!=)(<)]+)%s(\"[%a%d%s]+\")", "%1_%2_%3"), '_')
						if LootBuddy.db.char.dev_debug then
							print_str('pattern length: ' .. table.getn(pattern))
						end
						if table.getn(pattern) == 3 then
							pattern[3] = trim_str(pattern[3], "\"")
							if pattern[1] == "name" then
								filters[mi] = {}
								filters[mi][1] = pattern
								break
							else
								filters[mi][index] = pattern
							end
						else
							if LootBuddy.db.char.dev_debug then
								print_str('rule isnt correct')
							end
						end
					end
				end
			end
		--else
		--	if LootBuddy.db.char.dev_debug then
		--		print_str('filter string == nil')
		--	end
		--end
	else
		if LootBuddy.db.char.dev_debug then
			print_str('input == nil')
		end
	end
end

function LootBuddy:GetSafeList(info)
	return LootBuddy.db.char.safelist
end

function LootBuddy:SetSafeList(info, input)
	safe = {}
	LootBuddy.db.char.safelist = "" -- input
	if input ~= '' then
		local safeItems = string.explode(input, ",")
		--if table.getn(fstring) > 0 then
		for i,f in ipairs(safeItems) do
			if f ~= '' then
				table.insert(safe, f)
			end
		end
		LootBuddy.db.char.safelist = table.concat(safe, ", ")
	end
end

function print_str(s)
	DEFAULT_CHAT_FRAME:AddMessage('LootBuddy' .. s, 0.1, 0.7, 1.0);
end

function split_str(s, p, kv)
	t = {}
	for key,value in s:gmatch(p) do
		if kv == true then
			if value == nil then value = "nil" end
			t[key] = value
		else
			table.insert(t, key)
		end
	end
	if table.getn(t) == 0 then table.insert(t, s) end
	return t;
end

function string.explode(str, sep) 
	-- Function by Colandus
	local pos, t = 1, {}
	if #sep == 0 or #str == 0 then return end
	for s, e in function() return string.find(str, sep, pos) end do
		table.insert(t, string.trim(string.sub(str, pos, s-1)))
		pos = e+1
	end
	table.insert(t, string.trim(string.sub(str, pos)))
	return t
end

function trim_str(s, tchar)
  -- from PiL2 20.4
  return (s:gsub("^" .. tchar .. "*(.-)" .. tchar .. "*$", "%1"))
end

function in_table(t, sv)
	for _,v in pairs(t) do
		if v == sv then
			return true
		end
	end
	return false
end